package main_test

import (
	"encoding/json"
	"reflect"
	"testing"

	"github.com/pkg/errors"
	"github.com/stretchr/testify/assert"
)

type StringArray []string

func (sa *StringArray) UnmarshalJSON(data []byte) error {
	if len(data) > 0 && data[0] == '[' { // 配列
		var l []string
		if err := json.Unmarshal(data, &l); err != nil {
			return errors.Wrap(err, `failed to unmarshal array`)
		}
		*sa = StringArray(l)
		return nil
	}

	var s string
	if err := json.Unmarshal(data, &s); err != nil {
		return errors.Wrap(err, `failed to unmarshal string`)
	}
	*sa = StringArray([]string{s})
	return nil
}

func TestStringArray(t *testing.T) {
	const src1 = `"single element"`

	var sa StringArray
	if !assert.NoError(t, json.Unmarshal([]byte(src1), &sa), `json.Unmarshal should succeed`) {
		return
	}

	const src2 = `["one", "two", "three"]`
	if !assert.NoError(t, json.Unmarshal([]byte(src2), &sa), `json.Unmarshal should succeed`) {
		return
	}
}

func TestPerson(t *testing.T) {
	type Person struct {
		Name string `json:"name"`
		Age  int    `json:"age"`
	}

	const src = `{ "name": "lestrrat", "age": 41 }`
	var p Person
	if !assert.NoError(t, json.Unmarshal([]byte(src), &p), `jons.Unmarshal should succeed`) {
		return
	}
}

type PersonWithExtras1 struct {
	Name   string                 `json:"name"`
	Age    int                    `json:"age"`
	Extras map[string]interface{} `json:"-"`
}

func (p *PersonWithExtras1) UnmarshalJSON(data []byte) error {
	var m map[string]interface{}
	if err := json.Unmarshal(data, &m); err != nil {
		return errors.Wrap(err, `failed to unmsrhal json value`)
	}

	p.Name = m["name"].(string)
	p.Age = int(m["age"].(float64))
	delete(m, "name")
	delete(m, "age")
	p.Extras = m
	return nil
}

func TestPersonExtras1(t *testing.T) {
	const src = `{ "name": "lestrrat", "age": 41, "extra1": true, "extra2": "foo" }`
	var p PersonWithExtras1
	if !assert.NoError(t, json.Unmarshal([]byte(src), &p), `jons.Unmarshal should succeed`) {
		return
	}
}

type PersonWithExtras2 struct {
	Name   string                 `json:"name"`
	Age    int                    `json:"age"`
	Extras map[string]interface{} `json:"-"`
}

type personUnmarshalProxy struct {
	Name string `json:"name"`
	Age  int    `json:"age"`
}

func (p *PersonWithExtras2) UnmarshalJSON(data []byte) error {
	var proxy personUnmarshalProxy
	if err := json.Unmarshal(data, &proxy); err != nil {
		return errors.Wrap(err, `failed to unmarshal proxy`)
	}

	p.Name = proxy.Name
	p.Age = proxy.Age

	var m map[string]interface{}
	if err := json.Unmarshal(data, &m); err != nil {
		return errors.Wrap(err, `failed to unmarshal extras`)
	}
	p.Extras = m
	return nil
}

func TestPerson2(t *testing.T) {
	const src = `{ "name": "lestrrat", "age": 41, "extra1": true, "extra2": "foo" }`
	var p PersonWithExtras2
	if !assert.NoError(t, json.Unmarshal([]byte(src), &p), `jons.Unmarshal should succeed`) {
		return
	}
}

func makeProxyType(v interface{}) reflect.Type {
	typ := reflect.TypeOf(v)
	var fields []reflect.StructField
	for i := 0; i < typ.NumField(); i++ {
		fields = append(fields, typ.Field(i))
	}

	return reflect.StructOf(fields)
}

type PersonWithExtras3 struct {
	Name string `json:"name"`
	Age  int    `json:"age"`
}

var proxyT = makeProxyType(PersonWithExtras3{})

func (p *PersonWithExtras3) UnmarshalJSON(data []byte) error {
	proxy := reflect.New(proxyT).Interface()
	if err := json.Unmarshal(data, proxy); err != nil {
		return errors.Wrap(err, `failed to unmarshal`)
	}
	v := reflect.ValueOf(proxy).Elem()
	p.Name = v.FieldByName("Name").Interface().(string)
	p.Age = v.FieldByName("Age").Interface().(int)
	return nil
}

func TestPersonWithExtras3(t *testing.T) {
	var data = []string{
		`{ "name": "Alice", "age": 32 }`,
		`{ "name": "Bob",  "age": 35 }`,
	}

	for _, testcase := range data {
		var p PersonWithExtras3
		if !assert.NoError(t, json.Unmarshal([]byte(testcase), &p), `json.Unmarshal should succeed`) {
			return
		}
	}
}

type proxy struct {
	Type    string          `json:"type"`
	RawData json.RawMessage `json:"data"`
}

type Data struct {
	ObjectValue map[string]string
	ArrayValue  []string
}

func (d *Data) UnmarshalJSON(data []byte) error {
	// エラー処理は割愛してあります。
	var p proxy
	json.Unmarshal(data, &p)

	switch p.Type {
	case "object":
		var m map[string]string
		if err := json.Unmarshal(p.RawData, &m); err != nil {
			return errors.Wrap(err, `failed to unmarshal object value`)
		}
		d.ObjectValue = m
	case "array":
		var l []string
		if err := json.Unmarshal(p.RawData, &l); err != nil {
			return errors.Wrap(err, `failed to unmarshal array value`)
		}
		d.ArrayValue = l
	default:
		return errors.Errorf(`invalid type %v`, p.Type)
	}
	return nil
}

func TestDynamicParsing(t *testing.T) {
	var data = []string{
		`{ "type": "object", "data": { "name": "Alice" } }`,
		`{ "type": "array",  "data": [ "Bob", "Charlie" ] }`,
	}

	for _, testcase := range data {
		var d Data
		if !assert.NoError(t, json.Unmarshal([]byte(testcase), &d), `json.Unmarshal should succeed`) {
			return
		}
	}
}
